# Test if used
@testset "Used" begin
    # initialize model, parameter, and variable
    m = InfiniteModel()
    info = VariableInfo(false, 0, false, 0, false, 0, false, 0, false, false)
    v = build_variable(error, info, Hold)
    vref = add_variable(m, v, "name")
    # used_by_constraint
    @testset "used_by_constraint" begin
        # test not used
        @test !used_by_constraint(vref)
        # prepare use case
        m.var_to_constrs[JuMP.index(vref)] = [1]
        # test used
        @test used_by_constraint(vref)
        # undo changes
        delete!(m.var_to_constrs, JuMP.index(vref))
    end
    # used_by_measure
    @testset "used_by_measure" begin
        # test not used
        @test !used_by_measure(vref)
        # prepare use case
        m.var_to_meas[JuMP.index(vref)] = [1]
        # test used
        @test used_by_measure(vref)
        # undo changes
        delete!(m.var_to_meas, JuMP.index(vref))
    end
    # used_by_objective
    @testset "used_by_objective" begin
        # test not used
        @test !used_by_objective(vref)
        # prepare use case
        m.var_in_objective[JuMP.index(vref)] = true
        # test used
        @test used_by_objective(vref)
        # undo changes
        m.var_in_objective[JuMP.index(vref)] = false
    end
    # is_used
    @testset "is_used" begin
        # test not used
        @test !is_used(vref)
        # prepare use case
        m.var_to_constrs[JuMP.index(vref)] = [1]
        # test used
        @test is_used(vref)
        # undo changes
        delete!(m.var_to_constrs, JuMP.index(vref))
        # prepare use case
        m.var_to_meas[JuMP.index(vref)] = [1]
        # test used
        @test is_used(vref)
        # undo changes
        delete!(m.var_to_meas, JuMP.index(vref))
        # prepare use case
        m.var_in_objective[JuMP.index(vref)] = true
        # test used
        @test is_used(vref)
        # undo changes
        m.var_in_objective[JuMP.index(vref)] = false
    end
    # initialize infinite variable to test infinite methods
    param = InfOptParameter(IntervalSet(0, 1), Number[], false)
    pref = add_parameter(m, param, "test")
    v = build_variable(error, info, Infinite, parameter_refs = (pref, ))
    vref = add_variable(m, v, "name")
    # used_by_point_variable
    @testset "used_by_point_variable" begin
        # test not used
        @test !used_by_point_variable(vref)
        # build point variable
        v = build_variable(error, info, Point, infinite_variable_ref = vref,
                           parameter_values = 0.5)
        pvref = add_variable(m, v, "name2")
        # test used
        @test used_by_point_variable(vref)
    end
    # used_by_reduced_variable
    @testset "used_by_reduced_variable" begin
        # test not used
        @test !used_by_reduced_variable(vref)
        # build reduced variable
        m.infinite_to_reduced[JuMP.index(vref)] = [-1]
        # test used
        @test used_by_reduced_variable(vref)
    end
    # is_used (infinite variable)
    @testset "is_used (infinite var)" begin
        # test not used
        @test !is_used(vref)
        # prepare use case
        m.var_to_constrs[JuMP.index(vref)] = [1]
        # test used
        @test is_used(vref)
        # undo changes
        delete!(m.var_to_constrs, JuMP.index(vref))
        # prepare use case
        m.var_to_meas[JuMP.index(vref)] = [1]
        # test used
        @test is_used(vref)
        # undo changes
        delete!(m.var_to_meas, JuMP.index(vref))
        # prepare use case
        m.var_in_objective[m.next_var_index] = true
        # test used
        @test is_used(vref)
        # undo changes
        m.var_in_objective[m.next_var_index] = false
        # prepare use case
        m.reduced_to_constrs[-1] = [1]
        # test used
        @test is_used(vref)
        # undo changes
        delete!(m.reduced_to_constrs, -1)
        # prepare use case
        m.reduced_to_meas[-1] = [1]
        # test used
        @test is_used(vref)
        @test m.infinite_to_reduced[JuMP.index(vref)] == [-1]
    end
end

# Test variable queries for models
@testset "Model Queries" begin
    # initialize model, parameter, and variables
    m = InfiniteModel()
    param = InfOptParameter(IntervalSet(0, 1), Number[], false)
    pref = add_parameter(m, param, "test")
    info = VariableInfo(false, 0, false, 0, false, 0, false, 0, false, false)
    v = build_variable(error, info, Infinite, parameter_refs = (pref, ))
    ivref = add_variable(m, v, "name")
    v = build_variable(error, info, Point, infinite_variable_ref = ivref,
                       parameter_values = 0.5)
    pvref = add_variable(m, v, "name2")
    v = build_variable(error, info, Hold)
    hvref = add_variable(m, v, "name")
    # num_variables
    @testset "JuMP.num_variables" begin
        @test num_variables(m) == 3
    end
    # all_variables
    @testset "JuMP.all_variables" begin
        @test all_variables(m) == [ivref, pvref, hvref]
    end
end

# Test queries for parameter references and values
@testset "Parameter Tuples" begin
    # initialize model, parameter, and variables
    m = InfiniteModel()
    param = InfOptParameter(IntervalSet(0, 1), Number[], false)
    pref = add_parameter(m, param, "test")
    pref2 = add_parameter(m, param, "θ")
    info = VariableInfo(false, 0, false, 0, false, 0, false, 0, false, false)
    v = build_variable(error, info, Infinite, parameter_refs = (pref, ))
    ivref = add_variable(m, v, "name")
    ivref2 = add_variable(m, v, "name")
    v = build_variable(error, info, Point, infinite_variable_ref = ivref2,
                       parameter_values = 0.5)
    pvref = add_variable(m, v, "name2")
    # raw_parameter_values
    @testset "raw_parameter_values" begin
        @test raw_parameter_values(pvref) == VectorTuple{Float64}(0.5, )
    end
    # parameter_values
    @testset "parameter_values" begin
        @test parameter_values(pvref) == (Float64(0.5), )
    end
    # _update_variable_param_values
    @testset "_update_variable_param_values" begin
        @test isa(InfiniteOpt._update_variable_param_values(pvref, VectorTuple{Float64}(0)),
                  Nothing)
        @test parameter_values(pvref) == (Float64(0), )
    end
    # _update_variable_param_refs
    @testset "_update_variable_param_refs" begin
        @test isa(InfiniteOpt._update_variable_param_refs(ivref, VectorTuple(pref2)),
                  Nothing)
        @test parameter_refs(ivref) == (pref2, )
    end
    # set_parameter_refs
    @testset "set_parameter_refs" begin
        # test normal with 1 param
        @test isa(set_parameter_refs(ivref, (pref2, )), Nothing)
        @test parameter_refs(ivref) == (pref2, )
        @test name(ivref) == "name(θ)"
        # test double specify
        @test_throws ErrorException set_parameter_refs(ivref, (pref2, pref2))
        # test with used variable
        m.var_to_meas[JuMP.index(ivref)] = [1]
        @test isa(set_parameter_refs(ivref, (pref, )), Nothing)
        @test parameter_refs(ivref) == (pref, )
        @test name(ivref) == "name(test)"
        delete!(m.var_to_meas, JuMP.index(ivref))
        # test used by point variable
        m.infinite_to_points[JuMP.index(ivref)] = [1]
        @test_throws ErrorException set_parameter_refs(ivref, (pref, ))
        delete!(m.infinite_to_points, JuMP.index(ivref))
        # test used by reduced variable
        m.infinite_to_reduced[JuMP.index(ivref)] = [1]
        @test_throws ErrorException set_parameter_refs(ivref, (pref, ))
        delete!(m.infinite_to_reduced, JuMP.index(ivref))
    end
    # add_parameter_ref
    @testset "add_parameter_ref" begin
        # test used by point variable
        m.infinite_to_points[JuMP.index(ivref)] = [1]
        @test_throws ErrorException add_parameter_ref(ivref, pref2)
        delete!(m.infinite_to_points, JuMP.index(ivref))
        # test used by reduced variable
        m.infinite_to_reduced[JuMP.index(ivref)] = [1]
        @test_throws ErrorException add_parameter_ref(ivref, pref2)
        delete!(m.infinite_to_reduced, JuMP.index(ivref))
        # test normal use
        m.var_to_meas[JuMP.index(ivref)] = [1]
        @test isa(add_parameter_ref(ivref, pref2), Nothing)
        @test parameter_refs(ivref) == (pref, pref2)
        @test name(ivref) == "name(test, θ)"
        # test duplication error
        @test_throws ErrorException add_parameter_ref(ivref, pref2)
        #test bad array error
        @test_throws ErrorException add_parameter_ref(ivref, [pref2, pref])
        # reset
        delete!(m.var_to_meas, JuMP.index(ivref))
    end
end

# Test parameter bound queries for hold variables
@testset "Parameter Bounds" begin
    # initialize the model and other needed information
    m = InfiniteModel()
    @infinite_parameter(m, par in [0, 10])
    @infinite_parameter(m, pars[1:2] in [0, 10])
    @infinite_variable(m, inf(par))
    @hold_variable(m, x)
    @hold_variable(m, y, parameter_bounds = par == 0)
    data = DiscreteMeasureData(par, [1], [1])
    meas = measure(inf + par - x, data)
    # test parameter_bounds
    @testset "parameter_bounds" begin
        @test parameter_bounds(x) == ParameterBounds()
        @test parameter_bounds(y) == ParameterBounds(Dict(par => IntervalSet(0, 0)))
    end
    # test has_parameter_bounds
    @testset "has_parameter_bounds" begin
        @test !has_parameter_bounds(x)
        @test has_parameter_bounds(y)
        @test !has_parameter_bounds(InfiniteVariableRef(m, -2))
    end
    # test _update_variable_param_bounds
    @testset "_update_variable_param_bounds" begin
        dict = Dict(par => IntervalSet(0, 1))
        bounds = ParameterBounds(dict)
        @test isa(InfiniteOpt._update_variable_param_bounds(y, bounds), Nothing)
        @test parameter_bounds(y) == bounds
    end
    # test _check_meas_bounds (DiscreteMeasureData)
    @testset "_check_meas_bounds (DiscreteMeasureData)" begin
        data = DiscreteMeasureData(par, [1, 1], [1, 3], "", default_weight)
        # test ok
        bounds = ParameterBounds(Dict(par => IntervalSet(0, 5)))
        @test isa(InfiniteOpt._check_meas_bounds(bounds, data), Nothing)
        # test lower bad
        bounds = ParameterBounds(Dict(par => IntervalSet(2, 5)))
        @test_throws ErrorException InfiniteOpt._check_meas_bounds(bounds, data)
        # test upper bad
        bounds = ParameterBounds(Dict(par => IntervalSet(0, 2)))
        @test_throws ErrorException InfiniteOpt._check_meas_bounds(bounds, data)
    end
    # test _check_meas_bounds (MultiDiscreteMeasureData)
    @testset "_check_meas_bounds (Multi)" begin
        supps = [1 1; 3 3]
        data = MultiDiscreteMeasureData(pars, [0, 0], supps, "", default_weight)
        # test ok
        bounds = ParameterBounds(Dict(pars[1] => IntervalSet(0, 5)))
        @test isa(InfiniteOpt._check_meas_bounds(bounds, data), Nothing)
        # test lower bad
        bounds = ParameterBounds(Dict(pars[1] => IntervalSet(2, 5)))
        @test_throws ErrorException InfiniteOpt._check_meas_bounds(bounds, data)
        # test upper bad
        bounds = ParameterBounds(Dict(pars[2] => IntervalSet(0, 2)))
        @test_throws ErrorException InfiniteOpt._check_meas_bounds(bounds, data)
    end
    # test _check_meas_bounds (Fallback)
    @testset "_check_meas_bounds (Fallback)" begin
        warn = "Unable to check if hold variables bounds are valid in measure " *
               "with measure data type `BadData`. This can be resolved by " *
               "extending `measure_data_in_hold_bounds`."
        @test_logs (:warn, warn) InfiniteOpt._check_meas_bounds(ParameterBounds(),
                                                                BadData())
    end
    # test _update_bounds
    @testset "_update_bounds" begin
        # test normal
        dict1 = Dict(pars[1] => IntervalSet(0, 1), pars[2] => IntervalSet(-1, 2))
        dict2 = Dict(par => IntervalSet(0, 0), pars[1] => IntervalSet(0.5, 1.5),
                     pars[2] => IntervalSet(0, 1))
        @test isa(InfiniteOpt._update_bounds(dict1, dict2), Nothing)
        @test length(dict1) == 3
        @test dict1[par] == IntervalSet(0, 0)
        @test dict1[pars[1]] == IntervalSet(0.5, 1)
        @test dict1[pars[2]] == IntervalSet(0, 1)
        # test error
        new_dict = Dict(par => IntervalSet(1, 1))
        @test_throws ErrorException InfiniteOpt._update_bounds(dict1, new_dict)
        new_dict = Dict(par => IntervalSet(-1, -1))
        @test_throws ErrorException InfiniteOpt._update_bounds(dict1, new_dict)
    end
    # test _update_var_bounds (GeneralVariableRef)
    @testset "_update_var_bounds (General)" begin
        @test isa(InfiniteOpt._update_var_bounds(inf, ParameterBounds()),
                  Nothing)
    end
    # test _update_var_bounds (HoldVariableRef)
    @testset "_update_var_bounds (Hold)" begin
        bounds = ParameterBounds()
        InfiniteOpt._update_variable_param_bounds(x,
                                ParameterBounds(Dict(par => IntervalSet(1, 1))))
        @test isa(InfiniteOpt._update_var_bounds(x, bounds), Nothing)
        @test bounds.intervals[par] == IntervalSet(1, 1)
    end
    # test _update_var_bounds (MeasureRef)
    @testset "_update_var_bounds (Measure)" begin
        bounds = ParameterBounds()
        @test isa(InfiniteOpt._update_var_bounds(meas, bounds), Nothing)
        @test bounds.intervals[par] == IntervalSet(1, 1)
        InfiniteOpt._update_variable_param_bounds(x, ParameterBounds())
    end
    # test _rebuild_constr_bounds (BoundedScalarConstraint)
    @testset "_rebuild_constr_bounds (Bounded)" begin
        # test addition
        bounds = ParameterBounds(Dict(par => IntervalSet(0, 2)))
        c = BoundedScalarConstraint(y + inf, MOI.EqualTo(0.), copy(bounds),
                                    copy(bounds))
        expected = ParameterBounds(Dict(par => IntervalSet(0, 1)))
        @test InfiniteOpt._rebuild_constr_bounds(c, bounds).bounds == expected
        # test deletion
        bounds = ParameterBounds(Dict(par => IntervalSet(0, 0)))
        c = BoundedScalarConstraint(y + inf, MOI.EqualTo(0.), bounds,
                                    ParameterBounds())
        InfiniteOpt._update_variable_param_bounds(y, ParameterBounds())
        @test isa(InfiniteOpt._rebuild_constr_bounds(c, bounds),
                  ScalarConstraint)
        InfiniteOpt._update_variable_param_bounds(y, bounds)
    end
    # test _rebuild_constr_bounds (ScalarConstraint)
    @testset "_rebuild_constr_bounds (Scalar)" begin
        bounds = ParameterBounds(Dict(par => IntervalSet(0, 0)))
        c = ScalarConstraint(y + inf, MOI.EqualTo(0.))
        @test InfiniteOpt._rebuild_constr_bounds(c, bounds).bounds == bounds
    end
    # test set_parameter_bounds
    @testset "set_parameter_bounds" begin
        bounds1 = ParameterBounds(Dict(pars[1] => IntervalSet(0, 1)))
        # test force error
        @test_throws ErrorException set_parameter_bounds(y, bounds1)
        # prepare for main test
        data = DiscreteMeasureData(par, [1, 1], [1, 3], "", default_weight)
        mindex = -1; cindex1 = 42; cindex2 = -42; vindex = JuMP.index(x);
        m.measures[mindex] = Measure(x, data)
        m.var_to_meas[vindex] = [mindex]
        m.meas_to_constrs[mindex] = [cindex1]
        mref = MeasureRef(m, mindex)
        m.constrs[cindex1] = BoundedScalarConstraint(mref, MOI.EqualTo(0.0),
                                                     bounds1, bounds1)
        m.var_to_constrs[vindex] = [cindex2]
        m.constrs[cindex2] = JuMP.ScalarConstraint(x, MOI.EqualTo(0.0))
        # test measure error
        bounds = ParameterBounds(Dict(par => IntervalSet(1, 2)))
        @test_throws ErrorException set_parameter_bounds(x, bounds)
        # test constr error
        bounds = ParameterBounds(Dict(pars[1] => IntervalSet(2, 3)))
        @test_throws ErrorException set_parameter_bounds(x, bounds)
        InfiniteOpt._update_variable_param_bounds(x, ParameterBounds())
        # test normal
        bounds = ParameterBounds(Dict(par => IntervalSet(0, 5),
                                      pars[1] => IntervalSet(0.5, 2)))
        @test isa(set_parameter_bounds(x, bounds), Nothing)
        @test parameter_bounds(x) == bounds
        @test m.has_hold_bounds
        @test !optimizer_model_ready(m)
        expected = ParameterBounds(Dict(par => IntervalSet(0, 5),
                                        pars[1] => IntervalSet(0.5, 1)))
        @test m.constrs[cindex1].bounds == expected
        @test m.constrs[cindex2].bounds == bounds
        # Retest with empty constraint mapping
        delete!(m.var_to_constrs, vindex)
        @test isa(set_parameter_bounds(x, bounds, force = true), Nothing)
        @test parameter_bounds(x) == bounds
        @test m.has_hold_bounds
        @test !optimizer_model_ready(m)
        @test m.constrs[cindex1].bounds == expected
        m.var_to_constrs[vindex] = [cindex2]
        # test forced
        bounds = ParameterBounds(Dict(par => IntervalSet(1, 1)))
        @test isa(set_parameter_bounds(y, bounds, force = true), Nothing)
        @test parameter_bounds(y) == ParameterBounds(Dict(par => IntervalSet(1, 1)))
    end
    # test _update_constr_bounds (BoundedScalarConstraint)
    @testset "_update_constr_bounds (Bounded)" begin
        dict1 = Dict(pars[1] => IntervalSet(0, 1))
        dict2 = Dict(par => IntervalSet(0, 0), pars[1] => IntervalSet(1.1, 1.5))
        constr = BoundedScalarConstraint(par, MOI.EqualTo(0.0), ParameterBounds(copy(dict1)),
                                         ParameterBounds(copy(dict1)))
        # test error
        @test_throws ErrorException InfiniteOpt._update_constr_bounds(ParameterBounds(dict2), constr)
        @test constr.bounds == ParameterBounds(dict1)
        # test normal
        dict2 = Dict(par => IntervalSet(0, 0), pars[1] => IntervalSet(0.5, 1.5))
        dict = Dict(par => IntervalSet(0, 0), pars[1] => IntervalSet(0.5, 1))
        @test InfiniteOpt._update_constr_bounds(ParameterBounds(dict2), constr).bounds == ParameterBounds(dict)
    end
    # test _update_constr_bounds (ScalarConstraint)
    @testset "_update_constr_bounds (Scalar)" begin
        bounds = ParameterBounds(Dict(pars[1] => IntervalSet(0, 1)))
        constr = JuMP.ScalarConstraint(par, MOI.EqualTo(0.0))
        @test InfiniteOpt._update_constr_bounds(bounds, constr).bounds == bounds
    end
    # test add_parameter_bound
    @testset "add_parameter_bound" begin
        mindex = -1; cindex1 = 42; cindex2 = -42; vindex = JuMP.index(x);
        # test adding normally
        @test isa(add_parameter_bound(x, pars[2], 0, 5), Nothing)
        bounds = ParameterBounds(Dict(par => IntervalSet(0, 5),
                                      pars[1] => IntervalSet(0.5, 2),
                                      pars[2] => IntervalSet(0, 5)))
        @test parameter_bounds(x) == bounds
        expected = ParameterBounds(Dict(par => IntervalSet(0, 5),
                                        pars[1] => IntervalSet(0.5, 1),
                                        pars[2] => IntervalSet(0, 5)))
        @test m.constrs[cindex1].bounds == expected
        @test m.constrs[cindex2].bounds == bounds
        # test measure error
        @test_throws ErrorException add_parameter_bound(x, par, 1, 2)
        # test constr error
        @test_throws ErrorException add_parameter_bound(x, pars[1], 2, 3)
        # test overwritting existing
        @test isa(add_parameter_bound(y, par, 0, 0), Nothing)
        @test parameter_bounds(y) == ParameterBounds(Dict(par => IntervalSet(0, 0)))
    end
    # test delete_parameter_bound
    @testset "delete_parameter_bound" begin
        cindex1 = 42; cindex2 = -42;
        # test normal
        @test isa(delete_parameter_bound(x, pars[1]), Nothing)
        bounds = ParameterBounds(Dict(par => IntervalSet(0, 5),
                                      pars[2] => IntervalSet(0, 5)))
        @test parameter_bounds(x) == bounds
        expected = ParameterBounds(Dict(par => IntervalSet(0, 5),
                                        pars[1] => IntervalSet(0, 1),
                                        pars[2] => IntervalSet(0, 5)))
        @test m.constrs[cindex1].bounds == expected
        @test m.constrs[cindex2].bounds == bounds
        # test already gone
        @test isa(delete_parameter_bound(x, pars[1]), Nothing)
        @test parameter_bounds(x) == bounds
        @test m.constrs[cindex1].bounds == expected
        @test m.constrs[cindex2].bounds == bounds
    end
    # test delete_parameter_bounds
    @testset "delete_parameter_bounds" begin
        cindex1 = 42; cindex2 = -42;
        # test normal
        @test isa(delete_parameter_bounds(x), Nothing)
        @test parameter_bounds(x) == ParameterBounds()
        expected = ParameterBounds(Dict(pars[1] => IntervalSet(0, 1)))
        @test m.constrs[cindex1].bounds == expected
        @test isa(m.constrs[cindex2], ScalarConstraint)
        # test already gone
        @test isa(delete_parameter_bounds(x), Nothing)
        @test parameter_bounds(x) == ParameterBounds()
        @test m.constrs[cindex1].bounds == expected
        @test isa(m.constrs[cindex2], ScalarConstraint)
    end
end

# Test Parameter Bound Macros
@testset "Parameter Bound Macros" begin
    # initialize the model and other needed information
    m = InfiniteModel()
    @infinite_parameter(m, par in [0, 10])
    @infinite_parameter(m, pars[1:2] in [0, 10], container = SparseAxisArray)
    @hold_variable(m, x)
    @hold_variable(m, y, parameter_bounds = par == 0)
    # test @set_parameter_bounds
    @testset "@set_parameter_bounds" begin
        bounds1 = ParameterBounds(Dict(pars[1] => IntervalSet(0, 1)))
        # test force error
        @test_macro_throws ErrorException @set_parameter_bounds(y, pars[1] in [0, 1])
        # test bad arguments
        @test_macro_throws ErrorException @set_parameter_bounds(x, x, 23)
        @test_macro_throws ErrorException @set_parameter_bounds(x, pars[1] == 0,
                                                                bad = 42)
        # test bad input format
        @test_macro_throws ErrorException @set_parameter_bounds(x, pars[1] = 0)
        @test_macro_throws ErrorException @set_parameter_bounds(par, pars[1] == 0)
        # prepare for main test
        data = DiscreteMeasureData(par, [1, 1], [1, 3], "", default_weight)
        mindex = -1; cindex1 = 42; cindex2 = -42; vindex = JuMP.index(x);
        m.measures[mindex] = Measure(x, data)
        m.var_to_meas[JuMP.index(x)] = [mindex]
        m.meas_to_constrs[mindex] = [cindex1]
        mref = MeasureRef(m, mindex)
        m.constrs[cindex1] = BoundedScalarConstraint(mref, MOI.EqualTo(0.0),
                                                     bounds1, bounds1)
        m.var_to_constrs[vindex] = [cindex2]
        m.constrs[cindex2] = JuMP.ScalarConstraint(x, MOI.EqualTo(0.0))
        # test measure error
        @test_macro_throws ErrorException @set_parameter_bounds(x, par in [1, 2])
        # test constr error
        @test_macro_throws ErrorException @set_parameter_bounds(x, pars[1] in [2, 3])
        # test normal
        bounds = ParameterBounds(Dict(par => IntervalSet(0, 5),
                                      pars[1] => IntervalSet(0.5, 2)))
        @test isa(@set_parameter_bounds(x, (par in [0, 5], pars[1] in [0.5, 2])),
                  Nothing)
        @test parameter_bounds(x) == bounds
        @test m.has_hold_bounds
        @test !optimizer_model_ready(m)
        expected = ParameterBounds(Dict(par => IntervalSet(0, 5),
                                        pars[1] => IntervalSet(0.5, 1)))
        @test m.constrs[cindex1].bounds == expected
        @test m.constrs[cindex2].bounds == bounds
        # test forced
        @test isa(@set_parameter_bounds(y, par == 1, force = true), Nothing)
        @test parameter_bounds(y) == ParameterBounds(Dict(par => IntervalSet(1, 1)))
    end
    # test @add_parameter_bounds
    @testset "@add_parameter_bounds" begin
        mindex = -1; cindex1 = 42; cindex2 = -42; vindex = JuMP.index(x);
        # test adding normally
        @test isa(@add_parameter_bounds(x, pars[2] in [0, 5]), Nothing)
        dict = Dict(par => IntervalSet(0, 5), pars[1] => IntervalSet(0.5, 2),
                    pars[2] => IntervalSet(0, 5))
        @test parameter_bounds(x) == ParameterBounds(dict)
        expected = Dict(par => IntervalSet(0, 5), pars[1] => IntervalSet(0.5, 1),
                        pars[2] => IntervalSet(0, 5))
        @test m.constrs[cindex1].bounds == ParameterBounds(expected)
        @test m.constrs[cindex2].bounds == ParameterBounds(dict)
        # test measure error
        @test_macro_throws ErrorException @add_parameter_bounds(x, par in [1, 2])
        # test constr error
        @test_macro_throws ErrorException @add_parameter_bounds(x, pars[1] in [2, 3])
        # test bad input format
        @test_macro_throws ErrorException @add_parameter_bounds(x, pars[1] = 0)
        @test_macro_throws ErrorException @add_parameter_bounds(par, pars[1] == 0)
        # test overwritting existing
        @test isa(@add_parameter_bounds(y, par == 0), Nothing)
        @test parameter_bounds(y) == ParameterBounds(Dict(par => IntervalSet(0, 0)))
    end
end
